%{
// Copyright (C) 2009  Philip Ashmore (contact@philipashmore.com)
// License: LGPLv3.  See LICENSE.txt for the full license.

#include <iosfwd>
#include <fstream>
#include <stdexcept>
#include <sstream>
#include <ctype.h>
#include <vector>
#include <map>
#include <cstdlib> // system()
#include <cstring> // strcmp()
#include <v3c/string_utils.hpp>
#include <v3c/filesystem.hpp>

#ifdef V3C_FLEXX
#include <v3c/flexx/flexx.hpp>
#define ScannerBase v3c::Flexx
#define BEGIN begin
#else
#undef YY_DECL
#define YY_DECL int Scanner::read()
#define FLEXX_IMPL(S) public: S read()
#define ScannerBase yyFlexLexer
#endif

using namespace std;
using namespace v3c;

int n = 0;
enum { NFILES = 2 };
struct FileInfo
{
	int state;
	string name;
	const char * state_name;
	bool debug;
};
extern FileInfo fi[NFILES];
map<int, string> state_name;

string Class, semantic_value;
typedef vector< string > StringVec;
StringVec ns;
void extract_namespaces();

bool is_cxx = false, saw_noline = false;
string & cxx_file = fi[1].name;
string & flex_file = fi[0].name;
struct Define {
	inline Define(const string & key_, int value_)
	: key(key_), value(value_) {}
	string key;
	int value;
};
vector< Define > rules;
bool failed = false, saw_suffix_start = false, saw_yylex_while = false
, handled_yylex_while = false;

int table_lines = 0;
stringstream tables;

int num_transitions = 0;
vector<int> transitions;

class Scanner: public ScannerBase
{
	FLEXX_IMPL(int);
public:
	Scanner()
	: ScannerBase()
	, g_output_line()
	, state_eof_line(-10)
	{}
#ifndef V3C_FLEXX
	virtual void LexerError(const char* msg)
	{
		throw runtime_error(msg);
	}
#endif
	ostream & out() { return *yyout; }
	void add_lines(int n)
	{
		//int before = g_output_line / 5;
		g_output_line += n;
		//if(before != (g_output_line / 5))
		//	report_line();
	}
	void add_line()
	{
		add_lines(1);
	}
	void report_line()
	{
		//g_output_line++;
		//out() <<
		//"#line " << g_output_line << " // " << state_name[YY_START]
		//<<  '\n';
	}
	unsigned int g_output_line ,state_eof_line;
	void capture_rule(const char * p, unsigned int len)
	{
		string text(p, len);
		//cerr << "capture_rule \"" << text << "\"." << endl;
		stringstream ss(text);
		string key;
		int value;
		ss >> key >> value;
		if(!rules.size())
			out() << "{ " << key << '\n';
		else
			out() << ", " << key << '\n';
		add_line();
		rules.push_back(Define(key, value));
	}
	void transition(int x)
	{
#ifdef V3C_FLEXX
		out() << "/// " << m_state_names[x] << ", line " << g_output_line << endl;
		add_line();
		cerr << ">>> " << m_state_names[x] << '.' << endl;
#endif
		num_transitions++;
		transitions.push_back(x);
	}
	string namespace_open();
	string namespace_close();
	unsigned int ns_lines();
};
Scanner * pScanner = 0;
void process_options(const char * p, unsigned int len);
int process_line(const char * p, unsigned int len);
int num_rules = 0;
bool extract(const string & text, string & file, int & line)
{
	stringstream ss(text);
	char c;
	ss >> line >> c >> file;
	file = file.substr(0, file.length()-1); // Lose the trailing '"'
	if(!file.length())
		return false;
	cerr << "Line " << line << " \"" << file << "\"." << endl;
	return true;
}
%}

%option nounput batch yylineno
%option c++ 8bit warn noyywrap nodefault debug
%option yyclass="Scanner"
%x lex_flex lex_cpp lex_cpp_table lex_cpp_b4code lex_cpp_echo lex_cpp_code
%x lex_cpp_capture_rules lex_cpp_b4yylex lex_cpp_yylex lex_cpp_actions
%x lex_cpp_b4_suffix lex_cpp_suffix
%x lex_snip_debug lex_yy_get_next_buffer
letter    [abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_]
id        {letter}({letter}|[0-9])*
scoped_id {id}(::{id})*
space     [::space::]
%%
	set_debug(fi[n].debug);
	g_output_line = 1;
	cerr << "Entering state \"" << fi[n].state_name << "\"." << endl;
	if(n) { // .cpp
		pScanner->out() <<
			"// Generated by flex, modified by flexx.\n"
			"// This file is in the public domain.\n"
			"\n"
			;
		g_output_line += 3;
	}
	BEGIN(fi[n].state);
<lex_flex>{
^%option.+\n	process_options(yytext, yyleng);
.+	;
\n	;
}<lex_flex>

<lex_cpp>{
^"#define YY_NUM_RULES "[0-9]+\n	{
	string text(yytext + 21, yyleng - 21);
	stringstream ss(text);
	ss >> num_rules;
	cerr << "YY_NUM_RULES  defined as " << num_rules << endl;
}
^"static yyconst flex_int16_t yy_acclist["[0-9]+"] ="\n	{
	// This only appears if code contains REJECT.
	out() <<
	"#include <inttypes.h>\n"
	"#include <iostream>\n"
	"#include <cstdlib> // malloc\n"
	"#include <cerrno> // errno\n"
	"#include <string.h> // memset\n"
	"\n"
	"using namespace std;\n"
	"\n"
	"#ifdef V3C_FLEXX\n"
	"#error \"V3C_FLEXX should only be defined int the flexx implementation source file!\"\n"
	"#endif\n"
	"#define V3C_FLEXX 1\n"
	"\n"
	"#define YY_NUM_RULES " << num_rules << "\n"
	"#define YY_END_OF_BUFFER " << (num_rules+1) << "\n"
	"\n"
	"/* Action number for EOF rule of a given start state. */\n"
	"#define YY_STATE_EOF(state) (YY_END_OF_BUFFER + state + 1)\n"
	"#define YY_STATE_BUF_SIZE   ((YY_BUF_SIZE + 2) * sizeof(int))\n"
	"#define YY_END_OF_BUFFER_CHAR 0\n"
	"\n"
	;
	tables <<
	"namespace { // C++ way of saying \"static\".\n"
	"\n"
	"const int16_t " << "yy_acclist[] =\n"
	;
	add_lines(21);
	table_lines += 3;
	transition(lex_cpp_table);
	BEGIN(lex_cpp_table);
}
^"static yyconst flex_int16_t yy_accept["[0-9]+"] ="\n	{
	out() <<
	"#include <inttypes.h>\n"
	"#include <iostream>\n"
	"#include <cstdlib> // malloc\n"
	"#include <cerrno> // errno\n"
	"#include <string.h> // memset\n"
	"\n"
	"using namespace std;\n"
	"\n"
	"#ifdef V3C_FLEXX\n"
	"#error \"V3C_FLEXX should only be defined int the flexx implementation source file!\"\n"
	"#endif\n"
	"#define V3C_FLEXX 1\n"
	"\n"
	"#define YY_NUM_RULES " << num_rules << "\n"
	"#define YY_END_OF_BUFFER " << (num_rules+1) << "\n"
	"\n"
	"/* Action number for EOF rule of a given start state. */\n"
	"#define YY_STATE_EOF(state) (YY_END_OF_BUFFER + state + 1)\n"
	"#define YY_STATE_BUF_SIZE   ((YY_BUF_SIZE + 2) * sizeof(int))\n"
	"#define YY_END_OF_BUFFER_CHAR 0\n"
	"\n"
	;
	tables <<
	"namespace { // C++ way of saying \"static\".\n"
	"\n"
	"const int16_t " << "yy_accept[] =\n"
	;
	add_lines(21);
	table_lines += 3;
	transition(lex_cpp_table);
	BEGIN(lex_cpp_table);
}
^"/* Size of default input buffer. */"\n	{
	ECHO;
	add_line();
	transition(lex_cpp_echo);
	BEGIN(lex_cpp_echo);
}
.	;
\n	;
}<lex_cpp>
<lex_cpp_echo>{
^"/* The state buf must be large enough to hold one state per character in the main buffer." {
	transition(lex_cpp);
	BEGIN(lex_cpp);
}

yytext_ptr		out() << "yytext";
.	ECHO;
\n	ECHO;	add_line();
}<lex_cpp_echo>
<lex_cpp_table>{
^"static yyconst flex_int16_t yy_accept["[0-9]+"] ="\n	{
	tables <<
	"const int16_t " << "yy_accept[] =\n"
	;
	table_lines++;
}
^"static yyconst flex_int32_t yy_ec["[0-9]+"] ="\n	{
	tables <<
	"const int32_t " << "yy_ec[] =\n"
	;
	table_lines++;
}
^"static yyconst flex_int32_t yy_meta["[0-9]+"] ="\n	{
	tables <<
	"const int32_t " << "yy_meta[] =\n"
	;
	table_lines++;
}
^"static yyconst flex_int16_t yy_base["[0-9]+"] ="\n	{
	tables <<
	"const int16_t " << "yy_base[] =\n"
	;
	table_lines++;
}
^"static yyconst flex_int16_t yy_def["[0-9]+"] ="\n	{
	tables <<
	"const int16_t " << "yy_def[] =\n"
	;
	table_lines++;
}
^"static yyconst flex_int16_t yy_nxt["[0-9]+"] ="\n	{
	tables <<
	"const int16_t " << "yy_nxt[] =\n"
	;
	table_lines++;
}
^"static yyconst flex_int16_t yy_chk["[0-9]+"] ="\n	{
	tables <<
	"const int16_t " << "yy_chk[] =\n"
	;
	table_lines++;
}
^"static yyconst flex_int16_t yy_rule_linenum["[0-9]+"] ="\n	{
	tables <<
	"const int16_t " << "yy_rule_linenum[] =\n"
	;
	table_lines++;
}
^"static yyconst flex_int32_t yy_rule_can_match_eol["[0-9]+"] ="\n	{
	tables <<
	"const int32_t " << "yy_rule_can_match_eol[] =\n"
	;
	table_lines++;
}
^"#define yymore() ((yy_more_flag) = 1)"\n	|
^"#define yymore() yymore_used_but_not_detected"\n {
	transition(lex_cpp_b4code);
	BEGIN(lex_cpp_b4code);
}
.+	tables << string(yytext, yyleng);
\n	{
	tables << string(yytext, yyleng);
	table_lines++;
}
}<lex_cpp_table>

<lex_cpp_b4code>{
^#line\ [0-9]+\ \"[^\"]+\"\n	{
	string text(yytext + 6, yyleng - 7);
	cerr << "lex_cpp_b4code : " << text << endl;
	int line;
	string file;
	bool a = extract(text, file, line);
	bool b = (file == flex_file);
	bool c = (line == 2);
	//cerr << "extract(text, file, line) : " << a << endl;
	//cerr << file << " == " << flex_file << " : " << b << endl;
	//cerr << "line == 2 : " << c << endl;
	if(a && b && c) {
		cerr << "Start of code." << endl;
		transition(lex_cpp_code);
		BEGIN(lex_cpp_code);
	}
}
"{space}*FLEXX_IMPL{space}*({space}*{id}{space}*){space}*;{space}*"\n	{
	ECHO;
	if(!semantic_value.length()) {
		/// The semantic type may in general
		///  -# contain nested spaces (e.g. a templated type
		///  -# contain character literals
		///  -# contain '(' and ')'
		/// All we can do is count the '(' and ')' outside of strings and match
		/// them up, with the full type spec ending with the closing ')'.
		const char *p = yytext;
		while(*p && (*p != '(')) p++;
		if(!p) {
			cerr << "Couldn't find FLEXX_IMPL() open bracket" << endl;
			return 1;
		}
		int nest = 1;
		p++;
		while(*p && isspace(*p)) p++;
		if(!*p) {
			cerr << "Couldn't find FLEXX_IMPL() close bracket." << endl;
			return 1;
		}
		const char *s0 = p++;
		bool in_esc = false, in_string = false;
		for(; *p; p++) {
			if(in_esc) {
				in_esc = false;
				continue;
			}
			if(*p == '\\') {
				in_esc = true;
				continue;
			}
			if(*p == '"') {
				if(in_string)
					in_string = false;
				else
					in_string = true;
				continue;
			}
			if(*p == '(')
				nest++;
			else if(*p == ')') {
				nest--;
				if(!nest)
					break;
			}
		}
		if(nest) {
			cerr << "FLEXX_IMPL() bracket nesting incorrect." << endl;
			return 1;
		}
		p--;
		while(isspace(*p)) p--;
		semantic_value = string(s0, p - s0 + 1);
		if(!semantic_value.length()) {
			cerr << "FLEXX_IMPL() provides no semantic value." << endl;
			return 1;
		}
	} // if(!semantic_value.length())
}
.+	;
\n	;
}<lex_cpp_b4code>

<lex_cpp_code>{
^#line\ [0-9]+\ \"[^\"]+\"\n	{
	out() << tables.rdbuf() <<
	"\n"
	"} // namespace // C++ way of saying \"static\".\n"
	;
	add_lines(table_lines + 2);
	out() << namespace_open()
	<< "enum RuleEnum\n";
	;
	add_line();
	transition(lex_cpp_capture_rules);
	BEGIN(lex_cpp_capture_rules);
}
.+	ECHO;
\n	ECHO;	add_line();
}<lex_cpp_code>

<lex_cpp_capture_rules>{
^"#define "{id}" "[0-9]+\n	{
	capture_rule(yytext + 8, yyleng - 7);
}
^"#ifndef YY_NO_UNISTD_H"\n	{
	out() << "};\n";
	add_line();
	out() <<
	"const char * " << Class << "::m_state_names[] =\n";
	add_line();
	vector< Define >::const_iterator i = rules.begin(), i0 = i;
	vector< Define >::const_iterator j = rules.end();
	for(; i != j; ++i) {
		if(i == i0)
			out() << "{ ";
		else
			out() << ", ";
		out() << '"' << i->key << "\"\n";
		add_line();
	}
	out() << "};\n";
	add_line();
	transition(lex_cpp_b4yylex);
	BEGIN(lex_cpp_b4yylex);
}
.+	;
\n	;
}<lex_cpp_capture_rules>
<lex_cpp_b4yylex>{
^"YY_DECL"\n	{
	if(!semantic_value.length()) {
		//cerr << "FLEXX_IMPL() not found." << endl;
		//return 1;
		semantic_value = "int";
	}
	out() <<
	"/** TODO convert to debug stream.\n"
	" */\n"
	"void " << Class << "::report_state()\n"
	"{\n"
	"	if(yy_flex_debug) {\n"
	"		report_state_pre();\n"
	"		if( yy_act == 0)\n"
	"			cerr << \"--scanner backing up\";\n"
	"		else if ( yy_act < YY_NUM_RULES )\n"
	"			cerr << \"--accepting rule at line \"\n"
	"				<< yy_rule_linenum[yy_act] <<\n"
	"						\"(\\\"\" << yytext << \"\\\")\";\n"
	"		else if ( yy_act == YY_NUM_RULES )\n"
	"			cerr << \"--accepting default rule (\\\"\" << yytext << \"\\\")\";\n"
	"		else if ( yy_act == YY_END_OF_BUFFER )\n"
	"			cerr << \"--(end of buffer or a NUL)\";\n"
	"		else\n"
	"			cerr << \"--EOF (start condition \" << state_name()\n"
	"				 << \")\";\n"
	"		report_state_post();\n"
	"		cerr << '\\n';\n"
	"	}\n"
	"}\n"
	<< semantic_value << ' ' << Class << "::read()\n"
	//get yylex() to catch such that accept(num)/terminate() can be called from
	// anywhere.
	;
	add_lines(21);
	transition(lex_cpp_yylex);
	BEGIN(lex_cpp_yylex);
}
.+	;
\n	;
}<lex_cpp_b4yylex>
<lex_cpp_yylex>{
^"#ifdef YY_USER_INIT"\n	;
^"		YY_USER_INIT;"\n	out() << "\t\tinit();\n"; add_line();
^"#endif"\n	;
^"	while ( 1 )		/* loops until end-of-file is reached */"\n {
	ECHO;
	saw_yylex_while = true;
	add_line();
}
^"		{"\n	{
	if(saw_yylex_while && !handled_yylex_while) {
		ECHO;
		//out() << "\t\ttry{\n";
		handled_yylex_while = true;
	}else
		ECHO;
	add_line();
}
^"/* %% [12.0] debug code goes here */"\n	{
	ECHO;
	add_line();
	transition(lex_snip_debug);
	BEGIN(lex_snip_debug);
}
^"	{ /* beginning of action switch */"\n	{
	ECHO;
	add_line();
	transition(lex_cpp_actions);
	BEGIN(lex_cpp_actions);
}
^#line\ [0-9]+\ \"[^\"]+\"\n	{
	//ECHO;
	out() <<
	"#line " << (g_output_line + 1) << " \"" << cxx_file << "\"\n"
	;
	add_line();
}
^"	register int yy_act;"\n	;
^"	register yy_state_type yy_current_state;"\n ;
^"	register char *yy_cp, *yy_bp;"\n ;
yy_state_type	out() << "int";
yyalloc			out() << "malloc";
YY_FATAL_ERROR	out() << "LexerError";
yytext_ptr		out() << "yytext";
YY_CHAR			out() << "unsigned char";
YY_MORE_ADJ		out() << "yy_more_len";
YY_SC_TO_UI		out() << "promote";
YY_START		out() << "state()";
YY_NEW_FILE		out() << "yyrestart(yyin)";
yyalloc			out() << "malloc";
yylineno		out() << "line"; cerr << "line(yylineno)" << endl;
YY_CURRENT_BUFFER out() << "yy_CurrentBuffer()";
YY_CURRENT_BUFFER_LVALUE out() << "yy_CurrentBufferLvalue()";
YY_AT_BOL\(\)		out() << "yy_CurrentBufferLvalue()->yy_at_bol";
YY_DO_BEFORE_ACTION out() << "yy_do_before_action()";
.	ECHO;
\n	ECHO; add_line();
}<lex_cpp_yylex>
<lex_snip_debug>{
^"		switch ( yy_act )"\n	{
	out() <<
	"		if(yy_flex_debug)\n"
	"			report_state();\n"
	;
	ECHO;
	add_lines(3);
	transition(lex_cpp_yylex);
	BEGIN(lex_cpp_yylex);
}
.+	;
\n	;
}<lex_snip_debug>
<lex_cpp_actions>{
YY_RULE_SETUP	out() << "	{ RuleScope flexx_rule_scope(*this);";
YY_STATE_EOF\({id}\):	{
	state_eof_line = g_output_line;
	ECHO;
}
YY_BREAK		out() << "	} break;";
BEGIN			out() << "begin";
ECHO			out() << "echo()";
YYSTATE			out() << "state()";
YY_FATAL_ERROR	out() << "LexerError";
[ \t]+yyterminate\(\);\n	{
	if(state_eof_line == (g_output_line - 1)) {
		const char * p = yytext;
		while(isspace(*p)) p++;
		string indent(yytext, p - yytext);
		out() << indent << "{ RuleScope flexx_rule_scope(*this);\n"
			<< indent << "\tterminate();\n"
			<< indent << "}\n";
		add_lines(3);
	}else
		out() << "terminate()";
}
yytext_ptr		out() << "yytext";
YY_RESTORE_YY_MORE_OFFSET	;
YY_CURRENT_BUFFER out() << "yy_CurrentBuffer()";
YY_CURRENT_BUFFER_LVALUE out() << "yy_CurrentBufferLvalue()";
yy_state_type	out() << "int";
YY_MORE_ADJ		out() << "yy_more_len";
YY_SC_TO_UI		out() << "promote";
YY_START		out() << "state()";
YY_NEW_FILE		out() << "yyrestart(yyin)";
yyalloc			out() << "malloc";
yymore			out() << "more";
yyless			out() << "less";
yylineno		out() << "line"; cerr << "line(yylineno)" << endl;
^#line\ [0-9]+\ \"[^\"]+\"\n	{
	if(state_eof_line == (g_output_line - 1)) {
		out() << "	{ RuleScope flexx_rule_scope(*this);\n";
		add_line();
	}
	ECHO;
	add_line();
}
^"\t\t} /* end of scanning one token */"\n	{
	out() <<
	"\t\t}\n"
	;
	add_line();
	transition(lex_cpp_b4_suffix);
	BEGIN(lex_cpp_b4_suffix);
}
.	ECHO;
\n	ECHO; add_line();
}<lex_cpp_actions>
<lex_cpp_b4_suffix>{
^#line\ [0-9]+\ \".+\"\n	{
	string text(yytext + 6, yyleng - 7);
	int line;
	string file;
	if(extract(text, file, line) && file == flex_file) {
		cerr << "Start of suffix." << endl;
		//add_lines(2);
		transition(lex_cpp_suffix);
		out() << namespace_close();
		BEGIN(lex_cpp_suffix);
	}
	ECHO;
	add_line();
}
^"/* yy_get_next_buffer - try to read in a new buffer"\n {
	out() <<
	"}\n"
	;
	ECHO;
	add_lines(2);
	transition(lex_yy_get_next_buffer);
	BEGIN(lex_yy_get_next_buffer);
}
.+	;
\n	;
}<lex_cpp_b4_suffix>
<lex_yy_get_next_buffer>{
"yyFlexLexer"	out() << Class;
"yy_state_type"	out() << "int";
"yyunput"		out() << "unput";
"yyinput"		out() << "input";
YY_CURRENT_BUFFER out() << "yy_CurrentBuffer()";
YY_CURRENT_BUFFER_LVALUE out() << "yy_CurrentBufferLvalue()";
YY_FATAL_ERROR	out() << "LexerError";
yytext_ptr		out() << "yytext";
YY_CHAR			out() << "unsigned char";
YY_SC_TO_UI		out() << "promote";
YY_AT_BOL\(\)		out() << "yy_CurrentBufferLvalue()->yy_at_bol";
YY_MORE_ADJ		out() << "yy_more_len";
YY_NEW_FILE		out() << "yyrestart(yyin)";
YY_BUFFER_STATE	out() << "v3c::yy_buffer_state *";
yy_size_t		out() << "size_t";
yyalloc			out() << "malloc";
yyrealloc		out() << "realloc";
yyfree			out() << "free";
yylineno        out() << "line";
"struct yy_buffer_state"	out() << "v3c::yy_buffer_state";
YY_READ_BUF_SIZE	out() << "YY_BUF_SIZE";
YY_INPUT		out() << "check_input";
YY_START_STACK_INCR	out() << "start_stack_increment()";
YY_START		out() << "state()";
BEGIN			out() << "begin";
YY_START		out() << "state()";
yy_bs_lineno		out() << "line";
yy_bs_column		out() << "column";
^"void yyFlexLexer::LexerError( yyconst char msg[] )"\n {
	transition(lex_cpp_b4_suffix);
	BEGIN(lex_cpp_b4_suffix);
}
.	ECHO;
\n	ECHO;	add_line();
}<lex_yy_get_next_buffer>
<lex_cpp_suffix>{
.+	ECHO;
\n	{
	if(saw_suffix_start) {
		ECHO;
		add_line();
	}else
		saw_suffix_start = true;
}
}<lex_cpp_suffix>
%%
#define ISN(x) state_name[x] = #x
struct InitStateNames {
	InitStateNames()
	{
		ISN(INITIAL);
		ISN(lex_flex);
		ISN(lex_cpp);
		ISN(lex_cpp_table);
		ISN(lex_cpp_b4code);
		ISN(lex_cpp_echo);
		ISN(lex_cpp_code);
		ISN(lex_cpp_capture_rules);
		ISN(lex_cpp_b4yylex);
		ISN(lex_cpp_yylex);
		ISN(lex_cpp_actions);
		ISN(lex_cpp_b4_suffix);
		ISN(lex_cpp_suffix);
		ISN(lex_snip_debug);
		ISN(lex_yy_get_next_buffer);
	}
};
InitStateNames g_init_sn;

#define FI_ENTRY(name, debug) { name, "<?>", #name, debug }
FileInfo fi[NFILES] =
{ FI_ENTRY(lex_flex, false)
#ifdef V3C_FLEXX
, FI_ENTRY(lex_cpp, false)
#else
, FI_ENTRY(lex_cpp, false)
#endif
};

int process_file(int n_)
{
	n = n_;
	cerr << "Processing \"" << fi[n].name << "\"." << endl;
	ifstream in(fi[n].name.c_str());
	if(!in.good()) {
		cerr << "Couldn't open \"" << fi[n].name << "\" for reading." << endl;
		return 1;
	}
	string sout = string(fi[n].name);
	{
		string::size_type pos  = sout.rfind(DIR_SEP_CHAR);
		if(pos != string::npos)
			sout = sout.substr(pos + 1);
	}
	sout += ".out";
	ofstream out(sout.c_str());
	if(!out.good()) {
		cerr << "Couldn't open \"" << sout << "\" for writing." << endl;
		return 1;
	}
	Scanner scanner;
	pScanner = & scanner;
	scanner.switch_streams(&in, &out);
#ifdef V3C_FLEXX
	try {
		scanner.read();
	}catch(const v3c::scanner_result & e) {
		;
	}
#else
	scanner.read();
#endif
	if(failed) {
		cerr << "FAILED!" << endl;
		return 1;
	}
	return 0;
}
int main(int argc, char * argv[])
{
	if(argc < 2) {
		cerr << "Usage : flexx <flex options>" << endl;
		return 1;
	}
	flex_file = argv[argc - 1];
	string cmd = "flex";
	--argc, ++argv;
	bool add_last_arg = false;
	string arg;
	bool expect_value = false;
	for(; argc; --argc, ++argv) {
		if(add_last_arg)
			cmd += string(" ") + arg;
		else
			add_last_arg = true;
		arg = argv[0];
		if(expect_value) {
			cerr << "[command line] semantic value is \"" << arg << "\"."
				<< endl;
			semantic_value = arg;
			expect_value = false;
			add_last_arg = false;
			continue;
		}
		if(!is_cxx) {
			if(!strcmp(argv[0], "-+") || !strcmp(argv[0], "--c++")) {
				is_cxx = true;
				cerr << "[command line] %option c++ detected." << endl;
			}
		}
		if(arg == "-x") {
			cerr << "[command line] option -x detected." << endl;
			add_last_arg = false;
			expect_value = true;
		}
		if(arg == "-L" || arg == "--noline") {
			cerr << "flexx needs flex line numbers - ignoring \""
				<< arg << "\"." << endl;
			add_last_arg = false;
			continue;
		}
		string::size_type pos;
		pos = arg.find("--outfile=");
		if(pos != string::npos) {
			cxx_file = arg.substr(pos + 10);
			cerr << "[command line] %option outfile=\"" << cxx_file
				<< "\" detected." << endl;
			continue;
		}
		pos = arg.find("-o");
		if(pos != string::npos) {
			cxx_file = arg.substr(pos + 2);
			cerr << "[command line] %option outfile=\"" << cxx_file
				<< "\" detected." << endl;
			continue;
		}
		pos = arg.find("--yyclass=");
		if(pos != string::npos) {
			Class = arg.substr(pos + 10);
			cerr << "[command line] %option yyclass=\"" << Class
				<< "\" detected." << endl;
			extract_namespaces();
			continue;
		}
	}
	if(add_last_arg)
		cmd += string(" ") + arg;
	int ret = process_file(0);
	if(!is_cxx) {
		cerr << "flexx requires \"-+\" or \"--c++\" or \"%option c++\"."
			<< endl;
		return 1;
	}
	if(!Class.length()) {
		cerr << "flexx requires \"--yyclass=\" or \"%option yyclass=\"."
			<< endl;
		return 1;
	}
	if(!cxx_file.length()) {
		cerr << "flexx requires \"--outfile=\" or \"%option outfile\"."
			<< endl;
		return 1;
	}
	if(saw_noline) {
		cerr << "flexx needs flex line numbers - rejecting \"%option noline\"."
			<< endl;
		return 1;
	}
	cerr << "Running \"" << cmd << "\"." << endl;
	ret = system(cmd.c_str());
	if(ret) {
		cerr << "Flex exited with code " << ret << endl;
		return 1;
	}
	ret = process_file(1);
	if(ret)
		return ret;
	cerr << "Number of transitions : " << num_transitions << endl;
	cerr << state_name[INITIAL] << endl;
	// INITIAL -> lex_cpp_echo -> lex_cpp -> lex_cpp_table -> lex_cpp_b4code -> lex_cpp_code -> lex_cpp_capture_rules -> lex_cpp_b4yylex -> lex_cpp_yylex -> lex_snip_debug -> lex_cpp_yylex -> lex_cpp_actions -> lex_cpp_b4_suffix -> lex_yy_get_next_buffer -> lex_cpp_b4_suffix -> lex_cpp_suffix

	if(num_transitions != 15) {
		vector<int>::const_iterator i = transitions.begin();
		vector<int>::const_iterator j = transitions.end();
		for(; i != j; ++i)
			cerr << " -> " << state_name[*i];
		cerr << endl;
		cerr << "Flex generated unexpected code!" << endl;
		return 2;
	}
	cmd = "mv -f " + cxx_file + " " + cxx_file + ".orig";
	cerr << "Running \"" << cmd << "\"." << endl;
	if(system(cmd.c_str())) {
		cerr << "Command failed!" << endl;
		return 1;
	}

	cmd = "mv " + cxx_file + ".out " + cxx_file;
	cerr << "Running \"" << cmd << "\"." << endl;
	if(system(cmd.c_str())) {
		cerr << "Command failed!" << endl;
		return 1;
	}

	cerr << "Done." << endl;
	return 0;
}
void process_options(const char * p, unsigned int len)
{
	string text(p, len - 1); // \n
	cerr << "process_options \"" << text << "\"." << endl;
	//cerr << "Processing \"" << text << "\".";
	if(!is_cxx) {
		string::size_type pos = text.find(" c++ ");
		if(pos != string::npos) {
			is_cxx = true;
			cerr << "%option c++ detected." << endl;
		}
	}
	{
		string::size_type pos1 = text.find("noline");
		if(pos1 != string::npos)
			saw_noline = true;
	}
	do {
		// Look for outfile.
		string::size_type pos1 = text.find("outfile=\"");
		if(pos1 == string::npos)
			break;
		pos1 += 9; // skip to just past '"'.
		string::size_type pos2 = text.find('"', pos1);
		if(pos2 == string::npos) {
			cerr << "outfile has no close quote!" << endl;
			break;
		}
		if(pos2 == pos1 + 1) {
			cerr << "outfile string is empty!" << endl;
			break;
		}
		while(text[pos2 - 1] == '\\') {
			pos2 = text.find('"', pos2 + 1);
			if(pos2 == string::npos) {
				cerr << "outfile has no close quote!" << endl;
				break;
			}
		}
		cxx_file = text.substr(pos1, pos2 - pos1);
		cerr << "%option outfile=\"" << cxx_file << "\" detected." << endl;
	} while(false);
	do {
		// Look for %option yyclass="Scanner"
		string::size_type pos1 = text.find("yyclass=\"");
		if(pos1 == string::npos)
			break;
		pos1 += 9; // skip to just past '"'.
		string::size_type pos2 = text.find('"', pos1);
		if(pos2 == string::npos) {
			cerr << "yyclass has no close quote!" << endl;
			break;
		}
		if(pos2 == pos1 + 1) {
			cerr << "yyclass string is empty!" << endl;
			break;
		}
		while(text[pos2 - 1] == '\\') {
			pos2 = text.find('"', pos2 + 1);
			if(pos2 == string::npos) {
				cerr << "yyclass has no close quote!" << endl;
				break;
			}
		}
		Class = text.substr(pos1, pos2 - pos1);
		cerr << "%option yyclass=\"" << Class << "\" detected." << endl;
		extract_namespaces();
	} while(false);
}
void extract_namespaces()
{
	ns.clear();
	string::size_type pos = Class.find("::");
	while(pos != string::npos) {
		ns.push_back(Class.substr(0, pos));
		Class = Class.substr(pos + 2);
		pos = Class.find("::");
	}
}
string Scanner::namespace_open()
{
	string ret = string_inline_printf("/// namespace_open, line %u\n", g_output_line);
	StringVec::const_iterator i = ns.begin();
	StringVec::const_iterator j = ns.end();
	for(; i != j; ++i)
		ret += "namespace " + *i + "{\n";
	add_lines(ns_lines());
	return ret;
}
string Scanner::namespace_close()
{
	string ret = string_inline_printf("/// namespace_close, line %u\n", g_output_line);
	StringVec::const_reverse_iterator i = ns.rbegin();
	StringVec::const_reverse_iterator j = ns.rend();
	for(; i != j; ++i)
		ret += "} // namespace " + *i + "\n";
	add_lines(ns_lines());
	return ret;
}
unsigned int Scanner::ns_lines()
{
	return ns.size() + 1;
}
