%{
// Copyright (C) 2009  Philip Ashmore (contact@philipashmore.com)
// License: LGPLv3.  See LICENSE.txt for the full license.

#include <cstdlib>
#include <errno.h>
#include <limits.h>
#include <string>
#include "calc_parser.hpp"
%}

%option noyywrap nounput batch debug
%option outfile="calc_scanner.cpp"
%option c++ yyclass="calc::Scanner"

id    [a-zA-Z][a-zA-Z_0-9]*
int   [0-9]+
blank [ \t]

%%
{blank}+   |
\n+        |
\r+        ;
           /* Convert ints to the actual type of tokens.  */
[-+*/]     return make_token(yytext[0]);
":="       return make_token(ASSIGN);
{int}      return make_int_token();
{id}       return make_string_token();
.          error_char("invalid character");
%%
namespace calc {

TokenPtr Scanner::make_token(int id)
{
	return new Token(id, *this, yyleng);
}
TokenPtr Scanner::make_string_token()
{
	return new StringToken(*this, yytext, yyleng);
}
TokenPtr Scanner::make_int_token()
{
	long n = strtol(yytext, NULL, 10);
	if (! (INT_MIN <= n && n <= INT_MAX && errno != ERANGE))
		error("integer is out of range");
	return new IntToken(*this, yyleng, n);
}

} // namespace calc
