#if !defined(calc_scanner_hpp_included)
#define calc_scanner_hpp_included

// Copyright (C) 2009  Philip Ashmore (contact@philipashmore.com)
// License: LGPLv3.  See LICENSE.txt for the full license.

#include <stdexcept>
#include <sstream>
#include <v3c/flexx/flexx.hpp>
#include "calc_token.hpp"

namespace calc {

using namespace std;

class Scanner : public v3c::Flexx, public ParserTokens
{
	FLEXX_IMPL(TokenPtr);
public:
	Scanner(unsigned int spaces_per_tab = 4)
	: Flexx(spaces_per_tab)
	, ParserTokens()
	{}
	~Scanner() {}
	enum { eTerminateValue };
	virtual void error(const char* msg)
	{
		stringstream ss;
		ss << "At " << *this << " : " << msg << '.';
		throw runtime_error(ss.str());
	}
	virtual void error_char(const char* msg)
	{
		stringstream ss;
		ss << "At " << *this << " : " << msg << ' ';
		if(isprint(*yytext))
			ss << '\'' << *yytext << '\'';
		else
			ss << "0x" << hex << (unsigned int)(*yytext);
		ss << '.';
		throw runtime_error(ss.str());
	}
	virtual void before_rule();
	virtual void after_rule();
protected:
	TokenPtr make_token(int id);
	TokenPtr make_int_token();
	TokenPtr make_string_token();
private:
	Scanner(const Scanner & o);
	Scanner & operator =(const Scanner & o);
};

} // namespace calc

#endif // #if !defined(calc_scanner_hpp_included)
