#include "bisonx_grammar.hpp"

namespace v3c {
namespace bisonx {
namespace impl {

using namespace v3c;
using namespace std;

ostream & operator <<(ostream & s, const NodePtrVec & v)
{
	NodePtrVec::const_iterator i = v.begin();
	NodePtrVec::const_iterator j = v.end();
	for(unsigned int n = 0; i != j; ++i, ++n)
		s << '[' << n << "] " << *i << endl;
	return s;
}
PrettyPrintStreambuf & operator <<(PrettyPrintStreambuf & sb, const Token & t)
{
	ostream & s(sb.stream());
	t.stream_span(s);
	s << ", " << t.name() << " " << to_char_ptr(t.assoc()) << ", prec " << t.prec()
		<< ", number " << t.number();
	return sb;
}
/*PrettyPrintStreambuf & operator <<(PrettyPrintStreambuf & s, const SymbolPtr & t)
{
	if(t)
		s << *t;
	else
		s.stream() << "<null>";
	return s;
}*/
PrettyPrintStreambuf & operator <<(PrettyPrintStreambuf & sb, const Production & p)
{
	ostream & s(sb.stream());
	s << p.vec << endl;
	if(p.mergefunc.length())
		s << ", mergefunc " << p.mergefunc;
	s << ", assoc "
		<< to_char_ptr(p.assoc()) << ", prec " << p.prec() << ", dprec "
		<< p.dprec();
	return sb;
}
PrettyPrintStreambuf & operator <<(PrettyPrintStreambuf & sb, const ProductionVec & v)
{
	ostream & s(sb.stream());
	ProductionVec::const_iterator i = v.begin();
	ProductionVec::const_iterator j = v.end();
	for(unsigned int n = 0; i != j; ++i, ++n) {
		s << '[' << n << "] " << endl;
		sb.indent(true);
		sb << *i;
		sb.indent(false);
		s << endl;
	}
	return sb;
}
PrettyPrintStreambuf & operator <<(PrettyPrintStreambuf & sb, const Rule & r)
{
	sb.stream() << endl;
	sb.indent(true);
	sb << r.productions;
	sb.indent(false);
	return sb;
}
ostream & operator <<(ostream & s, const StringTokenPtr & t)
{
	if(t)
		s << *t;
	else
		s << "<null>";
	return s;
}
ostream & operator <<(ostream & s, const DefinesMap & m)
{
	DefinesMap::const_iterator i = m.begin();
	DefinesMap::const_iterator j = m.end();
	for(; i != j; ++i)
		s << i->first << " -> " << i->second << endl;
	return s;
}

ostream & operator <<(ostream & s, const Grammar & g)
{
	PrettyPrintStreambuf sb(s);
	s << "Grammar : start_symbol : " << g.start_symbol << ", default_prec : "
		<< g.default_prec << ", " << g.symbols.size() << " symbols." << endl;
	sb.indent(true);
	s << "prologue:" << endl;
	sb.indent(true);
	s << g.prologue << endl;
	sb.indent(false);
	s << "defines:" << endl;
	sb.indent(true);
	s << g.defines;
	sb.indent(false);
	s << "tokens:" << endl;
	sb.indent(true);
	//s << g.symbols;
	SymbolMap::const_iterator i = g.symbols.begin();
	SymbolMap::const_iterator j = g.symbols.end();
	for(; i != j; ++i)
		if(i->second->is_token()) {
			s << i->first << " -> ";
			sb << static_cast< const Token &>(*(i->second));
			s << endl;
		}
	sb.indent(false);
	s << "rules:" << endl;
	sb.indent(true);
	//s << g.symbols;
	i = g.symbols.begin();
	j = g.symbols.end();
	for(; i != j; ++i)
		if(i->second->is_rule()) {
			s << i->first << " -> ";
			sb << static_cast< const Rule &>(*(i->second));
			s << endl;
		}
	sb.indent(false);
	s << "epilogue:" << endl;
	sb.indent(true);
	s << g.epilogue << endl;
	sb.indent(false);
	return s;
}

} // namespace impl
} // namespace bisonx
} // namespace v3c
