#if !defined(v3c_parse_stack_hpp_included)
#define v3c_parse_stack_hpp_included

// Copyright (C) 2009  Philip Ashmore (contact@philipashmore.com)
// License: LGPLv3.  See LICENSE.txt for the full license.

#include <deque>

namespace v3c
{

template <class T, class S = std::deque<T> >
class parse_stack
{
public:

	// Hide our reversed order.
	typedef typename S::reverse_iterator iterator;
	typedef typename S::const_reverse_iterator const_iterator;

	parse_stack ()
	: seq_ ()
	{
	}
	parse_stack (unsigned int n) : seq_ (n)
	{
	}
	inline T& operator [] (unsigned int i)
	{
		return seq_[i];
	}
	inline const T& operator [] (unsigned int i) const
	{
		return seq_[i];
	}
	inline void push (const T& t)
	{
		seq_.push_front (t);
	}
	inline void pop (unsigned int n = 1)
	{
		for (; n; --n)
			seq_.pop_front ();
	}
	inline unsigned int size() const
	{
		return seq_.size ();
	}
	inline const_iterator begin () const { return seq_.rbegin (); }
	inline const_iterator end () const { return seq_.rend (); }
private:
	S seq_;
};

/// Present a slice of the top of a parse_stack.
template <class T, class S = parse_stack<T> >
class slice
{
public:
	slice (const S& parse_stack, unsigned int range)
	: stack_(parse_stack)
	, range_(range)
	{
	}
	inline const T& operator [] (unsigned int i) const
	{
		return stack_[range_ - i];
	}
private:
	const S& stack_;
	unsigned int range_;
};
} // namespace v3c

#endif // #if !defined(v3c_parse_stack_hpp_included)
